@@ -1,4 +1,4 @@
-/* modprobe.c: insert a module into the kernel, intelligently.
+/* modprobe.c: add or remove a module from the kernel, intelligently.
     Copyright (C) 2001  Rusty Russell.
     Copyright (C) 2002, 2003  Rusty Russell, IBM Corporation.
 
@@ -119,7 +119,7 @@ static void *do_nofail(void *ptr, const char *file, int line, const char *expr)
 static void print_usage(const char *progname)
 {
 	fprintf(stderr,
-		"Usage: %s [-v] [-V] [-C config-file] [-n] [-i] [-q] [-b] [-o <modname>] <modname> [parameters...]\n"
+		"Usage: %s [-v] [-V] [-C config-file] [-n] [-i] [-q] [-b] [-o <modname>] [ --dump-modversions ] <modname> [parameters...]\n"
 		"%s -r [-n] [-i] [-v] <modulename> ...\n"
 		"%s -l -t <dirname> [ -a <modulename> ...]\n",
 		progname, progname, progname);
@@ -270,7 +270,7 @@ static int add_modules_dep_line(char *line,
 	char *modname;
 
 	/* Ignore lines without : or which start with a # */
-	ptr = index(line, ':');
+	ptr = strchr(line, ':');
 	if (ptr == NULL || line[strspn(line, "\t ")] == '#')
 		return 0;
 
@@ -782,7 +782,7 @@ static void do_command(const char *modname,
 	while ((p = strstr(replaced_cmd, "$CMDLINE_OPTS")) != NULL) {
 		char *new;
 		asprintf(&new, "%.*s%s%s",
-			 p - replaced_cmd, replaced_cmd, cmdline_opts,
+			 (int)(p - replaced_cmd), replaced_cmd, cmdline_opts,
 			 p + strlen("$CMDLINE_OPTS"));
 		NOFAIL(new);
 		free(replaced_cmd);
@@ -830,7 +830,7 @@ static void insmod(struct list_head *list,
 	if (!list_empty(list)) {
 		insmod(list, NOFAIL(strdup("")), NULL, 0, warn,
 		       dry_run, verbose, options, commands, 0, ignore_proc,
-		       strip_vermagic, strip_modversion, cmdline_opts);
+		       strip_vermagic, strip_modversion, "");
 	}
 
 	/* Lock before we look, in case it's initializing. */
@@ -913,7 +913,8 @@ static void rmmod(struct list_head *list,
 		  struct module_command *commands,
 		  int ignore_commands,
 		  int ignore_inuse,
-		  const char *cmdline_opts)
+		  const char *cmdline_opts,
+		  int flags)
 {
 	const char *command;
 	unsigned int usecount = 0;
@@ -967,7 +968,7 @@ static void rmmod(struct list_head *list,
 	/* Now do things we depend. */
 	if (!list_empty(list))
 		rmmod(list, NULL, 0, warn, dry_run, verbose, commands,
-		      0, 1, cmdline_opts);
+		      0, 1, "", flags);
 	return;
 
 nonexistent_module:
@@ -990,13 +991,27 @@ static int type_matches(const char *path, const char *subpath)
 	return ret;
 }
 
+/* Careful!  Don't munge - in [ ] as per Debian Bug#350915 */
 static char *underscores(char *string)
 {
 	if (string) {
 		unsigned int i;
-		for (i = 0; string[i]; i++)
-			if (string[i] == '-')
-				string[i] = '_';
+		int inbracket = 0;
+		for (i = 0; string[i]; i++) {
+			switch (string[i]) {
+			case '[':
+				inbracket++;
+				break;
+			case ']':
+				inbracket--;
+				break;
+			case '-':
+				if (!inbracket)
+					string[i] = '_';
+			}
+		}
+		if (inbracket)
+			warn("Unmatched bracket in %s\n", string);
 	}
 	return string;
 }
@@ -1242,6 +1257,46 @@ static void read_toplevel_config(const char *filename,
 	}
 }
 
+/* Read possible module arguments from the kernel command line. */
+static int read_kcmdline(int dump_only, struct module_options **options)
+{
+	char *line;
+	unsigned int linenum = 0;
+	FILE *kcmdline;
+
+	kcmdline = fopen("/proc/cmdline", "r");
+	if (!kcmdline)
+		return 0;
+
+	while ((line = getline_wrapped(kcmdline, &linenum)) != NULL) {
+		char *ptr = line;
+		char *arg;
+
+		while ((arg = strsep_skipspace(&ptr, "\t ")) != NULL) {
+			char *sep, *modname, *opt;
+
+			sep = strchr(arg, '.');
+			if (sep) {
+				if (!strchr(sep, '='))
+					continue;
+				modname = arg;
+				*sep = '\0';
+				opt = ++sep;
+
+				if (dump_only)
+					printf("options %s %s\n", modname, opt);
+
+				*options = add_options(underscores(modname),
+						       opt, *options);
+			}
+		}
+
+		free(line);
+	}
+	fclose(kcmdline);
+	return 1;
+}
+
 static void add_to_env_var(const char *option)
 {
 	const char *oldenv;
@@ -1319,7 +1374,8 @@ static void handle_module(const char *modname,
 			  int strip_vermagic,
 			  int strip_modversion,
 			  int unknown_silent,
-			  const char *cmdline_opts)
+			  const char *cmdline_opts,
+			  int flags)
 {
 	if (list_empty(todo_list)) {
 		const char *command;
@@ -1341,7 +1397,7 @@ static void handle_module(const char *modname,
 
 	if (remove)
 		rmmod(todo_list, newname, first_time, error, dry_run, verbose,
-		      commands, ignore_commands, 0, cmdline_opts);
+		      commands, ignore_commands, 0, cmdline_opts, flags);
 	else
 		insmod(todo_list, NOFAIL(strdup(options)), newname,
 		       first_time, error, dry_run, verbose, modoptions,
@@ -1354,6 +1410,7 @@ static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { "config", 1, NULL, 'C' },
 				   { "name", 1, NULL, 'o' },
 				   { "remove", 0, NULL, 'r' },
+				   { "wait", 0, NULL, 'w' },
 				   { "showconfig", 0, NULL, 'c' },
 				   { "autoclean", 0, NULL, 'k' },
 				   { "quiet", 0, NULL, 'q' },
@@ -1416,6 +1473,7 @@ int main(int argc, char *argv[])
 	char *newname = NULL;
 	char *aliasfilename, *symfilename;
 	errfn_t error = fatal;
+	int flags = O_NONBLOCK|O_EXCL;
 
 	/* Prepend options from environment. */
 	argv = merge_args(getenv("MODPROBE_OPTIONS"), argv, &argc);
@@ -1430,7 +1488,7 @@ int main(int argc, char *argv[])
 		try_old_version("modprobe", argv);
 
 	uname(&buf);
-	while ((opt = getopt_long(argc, argv, "vVC:o:rknqQsclt:aifb", options, NULL)) != -1){
+	while ((opt = getopt_long(argc, argv, "vVC:o:rknqQsclt:aifbw", options, NULL)) != -1){
 		switch (opt) {
 		case 'v':
 			add_to_env_var("-v");
@@ -1515,6 +1573,9 @@ int main(int argc, char *argv[])
 		case 'b':
 			use_blacklist = 1;
 			break;
+		case 'w':
+			flags &= ~O_NONBLOCK;
+			break;
 		case 1:
 			strip_vermagic = 1;
 			break;
@@ -1565,6 +1626,7 @@ int main(int argc, char *argv[])
 
 		read_toplevel_config(config, "", 1, 0,
 			     &modoptions, &commands, &aliases, &blacklist);
+		read_kcmdline(1, &modoptions);
 		read_config(aliasfilename, "", 1, 0,&modoptions, &commands,
 			    &aliases, &blacklist);
 		read_config(symfilename, "", 1, 0, &modoptions, &commands,
@@ -1596,6 +1658,9 @@ int main(int argc, char *argv[])
 		read_toplevel_config(config, modulearg, 0,
 		     remove, &modoptions, &commands, &aliases, &blacklist);
 
+		/* Read module options from kernel command line */
+		read_kcmdline(0, &modoptions);
+
 		/* No luck?  Try symbol names, if starts with symbol:. */
 		if (!aliases
 		    && strncmp(modulearg, "symbol:", strlen("symbol:")) == 0)
@@ -1613,10 +1678,10 @@ int main(int argc, char *argv[])
 				read_config(aliasfilename, modulearg, 0,
 					    remove, &modoptions, &commands,
 					    &aliases, &blacklist);
-				aliases = apply_blacklist(aliases, blacklist);
 			}
 		}
 
+		aliases = apply_blacklist(aliases, blacklist);
 		if (aliases) {
 			errfn_t err = error;
 
@@ -1637,7 +1702,7 @@ int main(int argc, char *argv[])
 					      ignore_proc, strip_vermagic,
 					      strip_modversion,
 					      unknown_silent,
-					      optstring);
+					      optstring, flags);
 
 				aliases = aliases->next;
 				INIT_LIST_HEAD(&list);
@@ -1652,7 +1717,7 @@ int main(int argc, char *argv[])
 				      verbose, modoptions, commands,
 				      ignore_commands, ignore_proc,
 				      strip_vermagic, strip_modversion,
-				      unknown_silent, optstring);
+				      unknown_silent, optstring, flags);
 		}
 	}
 	if (log)
